# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import matplotlib.pyplot as plt

import burnman.tools

class SeismicData:
    """
    base class for all seismic models
    """
    def __init__(self):
        pass
        
    # depth in km
    def internal_depth_list(self):
        """ returns a sorted list of depths where this seismic data is computed at """
        return np.arange(0.,6000.0, 100.0)
        
    def evaluate_all_at(self, depth_list):
	""" returns pressure[Pa], density[kg/m^3], Vp[m/s], Vs[m/s] and Vphi[m/s] for a list of depths[m] """
        pressures = np.array([self.pressure(r) for r in depth_list])
        density = np.array([self.density(r) for r in depth_list])
        v_p = np.array([self.v_p(r) for r in depth_list])
        v_s = np.array([self.v_s(r) for r in depth_list])
        v_phi = np.array([self.v_phi(r) for r in depth_list])
        return pressures, density, v_p, v_s, v_phi

    def pressure(self, depth):
        raise ValueError, "not implemented"
        return 0

    def v_p(self, depth):
        raise ValueError, "not implemented"
        return 0

    def v_s(self, depth):
        raise ValueError, "not implemented"
        return 0
    
    def v_phi(self, depth):
        raise ValueError, "not implemented"
        return 0

    def density(self, depth):
        raise ValueError, "not implemented"
        return 0
    
    def depth(self, pressure):
        raise ValueError, "not implemented"
        return -1


class RadiusTable(SeismicData):
    """ 
    this is a base class that gets the information from a table indexed and
    sorted by radius. Fill the tables in the constructor after deriving
    from this class.
    Note: all tables need to be sorted by increasing radius.
    Alternatively, you can also overwrite the _lookup function if you
    want to access with something else like depth than radius.
    """ 
    def __init__(self):
        SeismicData.__init__(self)
        self.table_radius = []
        self.table_pressure = []
        self.table_density = []
        self.table_vp = []
        self.table_vs = []
        self.earth_radius = 6371.0e3
        
    def internal_depth_list(self):
        return (self.earth_radius - self.table_radius)[::-1] #radius is sorted in increasing order, so we need to reverse the depth list

    def pressure(self, depth):
        return self._lookup(depth, self.table_pressure)

    def v_p(self, depth):
        return self._lookup(depth, self.table_vp)

    def v_s(self, depth):
        return self._lookup(depth, self.table_vs)

    def v_phi(self, depth):
        v_s=self.v_s(depth)
        v_p=self.v_p(depth)
        return np.sqrt(v_p*v_p-4./3.*v_s*v_s)

    def density(self, depth):
        return self._lookup(depth, self.table_density)        

    def depth(self, pressure):
        radius = burnman.tools.lookup_and_interpolate(self.table_pressure[::-1], self.table_radius[::-1], pressure)
        return self.earth_radius - radius

    def _lookup(self, depth, value_table):
        radius = self.earth_radius - depth
        return burnman.tools.lookup_and_interpolate(self.table_radius, value_table, radius)    
    

class PREM(RadiusTable):
    """ 
    reads in the table for PREM (input_seismic/prem_table.txt) using the base class radiustable
    """
    def __init__(self):
        RadiusTable.__init__(self)
        table = burnman.tools.read_table("input_seismic/prem_table.txt") # radius, pressure, density, v_p, v_s
        table = np.array(table)
        self.table_radius = table[:,0]
        self.table_pressure = table[:,1] 
        self.table_density = table[:,2]
        self.table_vp = table[:,3]
        self.table_vs = table[:,4]

    def grav(self,depths):
        table = burnman.tools.read_table("input_seismic/grav_for_PREM.txt") # radius, g
        table = np.array(table)
        table_rad = table[:,0]
        table_g = table[:,1]
        return np.interp(self.earth_radius-depths, table_rad,table_g)


class Slow(RadiusTable):
    """ 
    Inserts the mean profiles for slower regions in the lower mantle (Lekic et al. 2012). 
    We need to stitch together three tables. Note that prem_lowermantle has a wider range, 
    so we cut away rows at the top and bottom. Interpolation is not necessary, 
    because all tables where generated with at the same depths 
    """
    def __init__(self):
        RadiusTable.__init__(self)

        table = burnman.tools.read_table("input_seismic/prem_lowermantle.txt")#data is: radius pressure density V_p V_s Q_K Q_G
        table = np.array(table)
        table[:,0] = table[:,0]  
        table2 = burnman.tools.read_table("input_seismic/swave_slow.txt")
        table2 = np.array(table2)
        table3 = burnman.tools.read_table("input_seismic/pwave_slow.txt")
        table3 = np.array(table3)

        min_radius = self.earth_radius-max(table2[:,0])
        max_radius = self.earth_radius-min(table2[:,0])
        
        table=np.array(filter(lambda x: (x[0]>=min_radius and x[0]<=max_radius), table))

        assert(len(table) == len(table2))
        assert(len(table) == len(table3))

        self.table_radius = table[:,0]
        self.table_pressure = table[:,1] 
        self.table_density = table[:,2]
        self.table_vp = table3[:,1] 
        self.table_vs = table2[:,1] 
       

class Fast(RadiusTable):
    """ 
    Inserts the mean profiles for faster regions in the lower mantle (Lekic et al. 2012). 
    We need to stitch together three tables. Note that prem_lowermantle has a wider range, 
    so we cut away rows at the top and bottom. Interpolation is not necessary, 
    because all tables where generated with at the same depths 
    """
    def __init__(self):
        RadiusTable.__init__(self)

        table = burnman.tools.read_table("input_seismic/prem_lowermantle.txt")#data is: radius pressure density V_p V_s Q_K Q_G
        table = np.array(table)
        table[:,0] = table[:,0]  
        table2 = burnman.tools.read_table("input_seismic/swave_fast.txt")
        table2 = np.array(table2)
        table3 = burnman.tools.read_table("input_seismic/pwave_fast.txt")
        table3 = np.array(table3)

        min_radius = self.earth_radius-max(table2[:,0])
        max_radius = self.earth_radius-min(table2[:,0])
        
        table=np.array(filter(lambda x: (x[0]>=min_radius and x[0]<=max_radius), table))

        assert(len(table) == len(table2))
        assert(len(table) == len(table3))

        self.table_radius = table[:,0]
        self.table_pressure = table[:,1] 
        self.table_density = table[:,2]
        self.table_vp = table3[:,1] 
        self.table_vs = table2[:,1] 



# this uses prem_lowermantle table
class PREMTest(RadiusTable):
    def __init__(self):
        RadiusTable.__init__(self)

        table = burnman.tools.read_table("input_seismic/prem_lowermantle.txt")
        #data is: radius pressure density V_p V_s Q_K Q_G
        table = np.array(table)
        self.table_radius = table[:,0]   
        self.table_pressure = table[:,1] 
        self.table_density = table[:,2]
        self.table_vp = table[:,3] 
        self.table_vs = table[:,4] 

def attenuation_correction(v_p,v_s,v_phi,Qs,Qphi):
    """ 
    Applies the attenuation correction following Matas et al. (2007), page 4. This is a minor effect on the velocities 
    """
    beta = 0.3 # Matas et al. (2007) page 4
    Qp  = 3./4.*pow((v_p/v_s),2.)*Qs    # Matas et al. (2007) page 4


    cot=1./np.tan(beta*np.pi/2.)
    v_p  = v_p*(1.-1./2.*cot*1./Qp)    # Matas et al. (2007) page 1
    v_s  = v_s*(1.-1./2.*cot*1./Qs)
    v_phi= v_phi*(1.-1./2.*cot*1./Qphi)
    return v_p, v_s, v_phi
    
# shared variable of prem, so that other routines do not need to create
# prem over and over. See geotherm for example.
prem_model = PREM()

if __name__ == "__main__":
    #create a seismic dataset from prem:
    s=PREM()
    depths = s.internal_depth_list()
    pressures, density, v_p, v_s, v_phi = s.evaluate_all_at(depths)
    print depths, pressures, density, v_p, v_s, v_phi


    # specify where we want to evaluate, here we map from pressure to depth, because we can
    #p = np.arange(1e9,360e9,5e9)
    #depths = map(s.depth, p) 
    #we could also just specify some depth levels directly like this:
    #depths = np.arange(35e3,5600e3,100e3)

    #now evaluate everything at the given depths levels (using interpolation)
    #pressures, density, v_p, v_s, v_phi = s.evaluate_all_at(depths)

    # plot vs and vp and v_phi (note that v_phi is computed!)
    plt.plot(depths/1.e3,v_p/1.e3,'+-r', label='v_p')
    plt.plot(depths/1.e3,v_s/1.e3,'+-b', label='v_s')
    plt.plot(depths/1.e3,v_phi/1.e3,'--g', label='v_phi')
    plt.legend()
    plt.xlabel('depth in km')
    plt.ylabel('km/s')
    plt.show()

    s1=PREM()
    depths=s1.internal_depth_list()
    pressures, density, v_p, v_s, v_phi = s1.evaluate_all_at(depths)
    plt.plot(depths/1.e3,v_p/1.e3,'+-r', label='v_p')
    plt.plot(depths/1.e3,v_s/1.e3,'+-b', label='v_s')
    plt.plot(depths/1.e3,v_phi/1.e3,'--g', label='v_phi')

    s2=prem_test()
    depths=s2.internal_depth_list()
    pressures, density, v_p, v_s, v_phi = s2.evaluate_all_at(depths)
    plt.plot(depths,v_p/1.e3,'x-r', label='v_p')
    plt.plot(depths,v_s/1.e3,'x-b', label='v_s')
    plt.plot(depths,v_phi/1.e3,'x-g', label='v_phi')
    plt.show()
    



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import warnings
from collections import namedtuple

from burnman.material import Material
from burnman.mineral import Mineral
    
def check_pairs(fractions, minerals):
        if len(fractions)<1:
            raise Exception('ERROR: we need at least one mineral')

        if len(fractions) != len(minerals):
            raise Exception('ERROR: different array lengths');

        total = sum(fractions)
        if abs(total-1.0)>1e-10:
            raise Exception('ERROR: list of molar fractions does not add up to one')
        for p in minerals:
            if not isinstance(p,Mineral):
                raise Exception('ERROR: object of type ''%s'' is not of type material' % (type(p)))


# static composite of minerals/composites
class Composite(Material):
    """
    Base class for a static composite material with fixed molar fractions. The
    elements can be minerals or materials, meaning composite can be nested
    arbitrarily.
    """
    def __init__(self, fractions, phases=None):
        """

        """

        if phases is None:
            # compatibility hack:
            tmp = fractions
            fractions = [pt[1] for pt in tmp]
            phases = [pt[0] for pt in tmp]

        assert(len(phases)==len(fractions))
        assert(len(phases)>0)
        for f in fractions:
            assert(f>=0)
        
        self.children = zip(fractions, phases)

        total = sum(fractions)

        if (total-1.0)>1e-8:
            warnings.warn('Warning: list of molar fractions does not add up to one. Normalizing.')
            for (fraction,phase) in self.children:
                fraction = fraction / total


    def set_method(self, method):
        """
        set the same equation of state method for all the phases in the composite
        """
        for (fraction, phase) in self.children:
            phase.set_method(method) 

    def unroll(self):
        fractions = []
        minerals = []

        for (fraction, phase) in self.children:
            p_fr,p_min = phase.unroll()
            check_pairs(p_fr, p_min)
            fractions.extend([i*fraction for i in p_fr])
            minerals.extend(p_min)
        return (fractions, minerals)

    def to_string(self):
        """
        return the name of the composite
        """
        return "'" + self.__class__.__name__ + "'"

    def debug_print(self):
        (fr,mins) = self.unroll()
        for (fr,mi) in zip(fr,mins):
            print "%g of phase %s"%(fr,mi.to_string())

    def set_state(self, pressure, temperature):
        """
        Update the material to the given pressure [Pa] and temperature [K].
        """
        self.pressure = pressure
        self.temperature = temperature
        for (fraction, phase) in self.children:
            phase.set_state(pressure, temperature) 

    def density(self):
        """
        Compute the density of the composite based on the molar volumes and masses
        """
        densities = np.array([ph.density() for (_,ph) in self.children])
        volumes = np.array([ph.molar_volume()*fraction for (fraction, ph) in self.children])
        return np.sum(densities*volumes)/np.sum(volumes)




# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

class EquationOfState:
    """
    This class defines the interface for an equation of state
    that a mineral uses to determine its properties at a 
    given P,T.  In order define a new equation of state, you
    should define these functions.

    All functions should accept and return values in SI units.
   
    In general these functions are functions of pressure,
    temperature, and volume, as well as a "params" object,
    which stores the material parameters of the stuff,
    such as reference volume, Debye temperature, etc. 
    The exceptions are volume and density, which are
    just assumed to be functions of pressure and temperature.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns molar volume at the pressure and temperature [m^3]
        """
        raise NotImplementedError("")

    def density(self, pressure, temperature, params):
        """
        Returns density at the pressure and temperature [kg/m^3]
        """
        return params["molar_mass"] / self.volume(pressure, temperature, params)

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter at the pressure, temperature, and volume
        """
        raise NotImplementedError("")

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus at the pressure, temperature, and volume [Pa]
        """
        raise NotImplementedError("")

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus at the pressure, temperature, and volume [Pa]
        """
        raise NotImplementedError("")

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus at the pressure, temperature, and volume [Pa]
        """
        raise NotImplementedError("")

    def heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature, and volume [J/K/mol]
        """
        raise NotImplementedError("")

    def heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure at the pressure, temperature, and volume [J/K/mol]
        """
        raise NotImplementedError("")

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature, and volume [1/K]
        """
        raise NotImplementedError("")




# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
SLB_2005
^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2005 and references therein 

"""
import burnman.mineral_helpers as bmb
from burnman.mineral import Mineral
                

class stishovite (Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2005 and references therein 
    """
    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 14.02e-6,
            'K_0': 314.0e9,
            'Kprime_0': 4.4,
            'G_0': 220.0e9,
            'Gprime_0': 1.6,
            'molar_mass': .0601,
            'n': 3,
            'Debye_0': 1044.,
            'grueneisen_0': 1.34,
            'q_0': 2.4,
            'eta_s_0': 5.0 }

class periclase (Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2005 and references therein 
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 11.24e-6,
            'K_0': 161.0e9,
            'Kprime_0': 3.8,
            'G_0': 131.0e9,
            'Gprime_0': 2.1,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 773.,
            'grueneisen_0': 1.5,
            'q_0': 1.5,
            'eta_s_0': 2.8 }

class wuestite (Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2005 and references therein 
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 12.06e-6,
            'K_0': 152.0e9,
            'Kprime_0': 4.9,
            'G_0': 47.0e9,
            'Gprime_0': 0.7,
            'molar_mass': .0718,
            'n': 2,
            'Debye_0': 455.,
            'grueneisen_0': 1.28,
            'q_0': 1.5,
            'eta_s_0': 0.8 }



class ferropericlase(bmb.HelperSolidSolution):
    def __init__(self, fe_num):
        base_materials = [periclase(), wuestite()]
        molar_fraction = [1. - fe_num, 0.0 + fe_num] # keep the 0.0 +, otherwise it is an array sometimes
        bmb.HelperSolidSolution.__init__(self, base_materials, molar_fraction)



class mg_fe_perovskite(bmb.HelperSolidSolution):
    def __init__(self, fe_num):
        base_materials = [mg_perovskite(), fe_perovskite()]
        molar_fraction = [1. - fe_num, 0.0 + fe_num] # keep the 0.0 +, otherwise it is an array sometimes
        bmb.HelperSolidSolution.__init__(self, base_materials, molar_fraction)


class mg_perovskite(Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2005 and references therein  
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 24.45e-6,
            'K_0': 251.0e9,   
            'Kprime_0': 4.1,     
            'G_0': 175.0e9,  
            'Gprime_0': 1.7,  
            'molar_mass': .1000,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.6 }

class fe_perovskite(Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2005 and references therein 
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 25.48e-6,
            'K_0': 281.0e9, 
            'Kprime_0': 4.1,  
            'G_0': 138.0e9,
            'Gprime_0': 1.7,   
            'molar_mass': .1319, 
            'n': 5,
            'Debye_0': 841.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.1 }


class mg_fe_perovskite_pt_dependent(bmb.HelperFeDependent):
    def __init__(self, iron_number_with_pt, idx):
        bmb.HelperFeDependent.__init__(self, iron_number_with_pt, idx)

    def create_inner_material(self, iron_number):
        return mg_fe_perovskite(iron_number)

class ferropericlase_pt_dependent(bmb.HelperFeDependent):
    def __init__(self, iron_number_with_pt, idx):
        bmb.HelperFeDependent.__init__(self, iron_number_with_pt, idx)

    def create_inner_material(self, iron_number):
        return ferropericlase(iron_number)



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np

# TODO: add up weight percent and check <100 and tell them how much

molar_mass = {'Fe':55.845/1000., 'Mg':24.305/1000., 'O':15.999/1000., 'Al':26.982/1000., 'Ca':40.078/1000., 'Si':28.085/1000.} # kg/mol
Av = 6.022141e23 # Avogadro constant in 1/mol 
boltzmann_constant = 1.3806503e-23 # in m^2 kg s^-2 K^-1
gas_constant = Av * boltzmann_constant # in J mol^-1 K^-1

lower_mantle_mass = 4.043e24*.75 # in kg




# convert weight percentage (amount, 1.00 = 100%) of a given element to molar mass
def weight_pct_to_mol(element, amount):

    return amount * lower_mantle_mass / molar_mass[element] * Av


def calculate_phase_percents(inp):
    """
    Converts given weight percentages into the requisite percent of each phase
    in mols and also returns the fraction of perovskite versus ferropericlase, 
    assuming all of the silcon goes into the perovskite phase
    and with any remaining Fe or Mg going into the oxide phase.
    Input:
    inp={'Mg': ..., 'Fe': ..., ...} # in weight percent
    Returns:
    phase_per={'fp': ..., 'pv': ...} # as a fraction
    rel_mol_per={'MgO: ..., 'FeO': ..., ...} # in mols 
    """
    names = {'Mg':'MgO','Fe':'FeO','Si':'SiO2', 'Ca':'Ca', 'Al':'Al'}
    rel_mol_per = {}
    out = {}
    for a in inp:
        out[names[a]] = weight_pct_to_mol(a,inp[a])


    norm = out['MgO']+out['FeO']
    for a in inp:
        rel_mol_per[names[a]] = out[names[a]]/norm

    frac_mol_SiO2 = rel_mol_per['SiO2']
    phase_per={'fp':(1.-frac_mol_SiO2),'pv':frac_mol_SiO2}
    return phase_per,rel_mol_per
    
    
    
def part_coef_calc(inp2,StartP,EndP,deltaP):

    raise Exception("not implemented")

    a = [] #partition coefficent of Fe in fp
    b = [] #partition coefficent of Fe in pv
    
    Pressure= []
    Temperature=[]
    counter = 0

    
    
   
def calculate_partition_coefficient(pressure, temperature, components, initial_distribution_coefficient):

    """ calculate the partition coefficient given [...] initial_distribution_coefficient is known as Kd_0 """

    frac_mol_FeO = components['FeO']
    frac_mol_SiO2 = components['SiO2']
    Kd_0 = initial_distribution_coefficient

    delV = 2.e-7 #in m^3/mol, average taken from Nakajima et al 2012, JGR
    

    rs = ((25.e9-pressure)*(delV)/(gas_constant*temperature))+np.log(Kd_0) #eq 5 Nakajima et al 2012

    K = np.exp(rs) #The exchange coefficent at P and T

    num_to_sqrt = (-4.*frac_mol_FeO*(K-1.)*K*frac_mol_SiO2)+(pow(1.+(frac_mol_FeO*(K-1))+((K-1.)*frac_mol_SiO2),2.))

    b = (-1. + frac_mol_FeO - (frac_mol_FeO*K)+frac_mol_SiO2 - (frac_mol_SiO2*K) + np.sqrt(num_to_sqrt)) \
         / (2.*frac_mol_SiO2*(1.-K))

    a = b /(((1.-b)*K)+b)
    
    return (a,b) #a is partition coefficient array with P for mw, b is pcarray for pv




# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import scipy.optimize as opt
import warnings

import burnman.birch_murnaghan as bm
import burnman.debye as debye
import burnman.equation_of_state as eos
 
class SLBBase(eos.EquationOfState):
    """
    Base class for the finite strain-Mie-Grueneiesen-Debye equation of state detailed
    in Stixrude and Lithgow-Bertelloni (2005).  For the most part, the equations are 
    all third order in strain, but see further the slb2 and slb3 classes
    """
    def __debye_temperature(self,x,params):
        """
        Finite strain approximation for Debye Temperature [K]
        x = ref_vol/vol
        """
        f = 1./2. * (pow(x, 2./3.) - 1.)
        a1_ii = 6. * params['grueneisen_0'] # EQ 47
        a2_iikk = -12.*params['grueneisen_0']+36.*pow(params['grueneisen_0'],2.) - 18.*params['q_0']*params['grueneisen_0'] # EQ 47
        return params['Debye_0'] * np.sqrt(1. + a1_ii * f + 1./2. * a2_iikk*f*f) 

    def volume_dependent_q(self, x, params):
        """
        Finite strain approximation for q, the isotropic volume strain
        derivative of the grueneisen parameter
        """
        f = 1./2. * (pow(x, 2./3.) - 1.)
        a1_ii = 6. * params['grueneisen_0'] # EQ 47
        a2_iikk = -12.*params['grueneisen_0']+36.*pow(params['grueneisen_0'],2.) - 18.*params['q_0']*params['grueneisen_0'] # EQ 47
        nu_o_nu0_sq = 1.+ a1_ii*f + (1./2.)*a2_iikk * f*f # EQ 41
        gr = 1./6./nu_o_nu0_sq * (2.*f+1.) * ( a1_ii + a2_iikk*f )
        q = 1./9.*(18.*gr - 6. - 1./2. / nu_o_nu0_sq * (2.*f+1.)*(2.*f+1.)*a2_iikk/gr)
        return q
    
    def __isotropic_eta_s(self, x, params):
        """
        Finite strain approximation for eta_s_0, the isotropic shear
        strain derivative of the grueneisen parameter
        """
        f = 1./2. * (pow(x, 2./3.) - 1.)
        a2_s = -2.*params['grueneisen_0'] - 2.*params['eta_s_0'] # EQ 47 
        a1_ii = 6. * params['grueneisen_0'] # EQ 47
        a2_iikk = -12.*params['grueneisen_0']+36.*pow(params['grueneisen_0'],2.) - 18.*params['q_0']*params['grueneisen_0'] # EQ 47
        nu_o_nu0_sq = 1.+ a1_ii*f + (1./2.)*a2_iikk * pow(f,2.) # EQ 41
        gr = 1./6./nu_o_nu0_sq * (2.*f+1.) * ( a1_ii + a2_iikk*f )
        eta_s = - gr - (1./2. * pow(nu_o_nu0_sq,-1.) * pow((2.*f)+1.,2.)*a2_s) # EQ 46 NOTE the typo from Stixrude 2005
        return eta_s


    def volume(self, pressure, temperature, params):
        """
        Returns molar volume at the pressure and temperature [m^3]
        """
        debye_T = lambda x : self.__debye_temperature(params['V_0']/x, params)
        gr = lambda x : self.grueneisen_parameter(pressure, temperature, x, params)
        E_th =  lambda x : debye.thermal_energy(temperature, debye_T(x), params['n']) #thermal energy at temperature T
        E_th_ref = lambda x : debye.thermal_energy(300., debye_T(x), params['n']) #thermal energy at reference temperature
      
        b_iikk= 9.*params['K_0'] # EQ 28
        b_iikkmm= 27.*params['K_0']*(params['Kprime_0']-4.) # EQ 29
        f = lambda x: 0.5*(pow(params['V_0']/x,2./3.)-1.) # EQ 24
        func = lambda x: (1./3.)*(pow(1.+2.*f(x),5./2.))*((b_iikk*f(x)) \
            +(0.5*b_iikkmm*pow(f(x),2.))) + gr(x)*(E_th(x) - E_th_ref(x))/x - pressure #EQ 21 

        # we need to have a sign change in [a,b] to find a zero. Let us start with a
        # conservative guess:
        a = 0.6*params['V_0']
        b = 1.2*params['V_0']

        # if we have a sign change, we are done:
        if func(a)*func(b)<0:
            return opt.brentq(func, a, b) 
        else:
            tol = 0.0001
            sol = opt.fmin(lambda x : func(x)*func(x), 1.0*params['V_0'], ftol=tol, full_output=1, disp=0)
            if sol[1] > tol*2:
                raise ValueError('Cannot find volume, likely outside of the range of validity for EOS')
            else:
                warnings.warn("May be outside the range of validity for EOS")
                return sol[0]

    

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter at the pressure, temperature, and volume
        """
        x = params['V_0'] / volume
        f = 1./2. * (pow(x, 2./3.) - 1.)
        gruen_0 = params['grueneisen_0']
        a1_ii = 6. * gruen_0 # EQ 47
        a2_iikk = -12.*gruen_0 + 36.*gruen_0*gruen_0 - 18.*params['q_0']*gruen_0 # EQ 47
        nu_o_nu0_sq = 1.+ a1_ii*f + (1./2.)*a2_iikk * f*f # EQ 41
        return 1./6./nu_o_nu0_sq * (2.*f+1.) * ( a1_ii + a2_iikk*f )

    def isothermal_bulk_modulus(self, pressure,temperature, volume, params):
        """
        Returns isothermal bulk modulus at the pressure, temperature, and volume [Pa]
        """
        debye_T = self.__debye_temperature(params['V_0']/volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)

        E_th = debye.thermal_energy(temperature, debye_T, params['n']) #thermal energy at temperature T
        E_th_ref = debye.thermal_energy(300.,debye_T, params['n']) #thermal energy at reference temperature
    
        C_v = debye.heat_capacity_v(temperature, debye_T, params['n']) #heat capacity at temperature T
        C_v_ref = debye.heat_capacity_v(300.,debye_T, params['n']) #heat capacity at reference temperature

        q = self.volume_dependent_q(params['V_0']/volume, params)
    
        K = bm.bulk_modulus(volume, params) \
            + (gr + 1.-q)* ( gr / volume ) * (E_th - E_th_ref) \
            - ( pow(gr , 2.) / volume )*(C_v*temperature - C_v_ref*300.)   
    
        return K
    
    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus at the pressure, temperature, and volume [Pa]
        """
        K_T=self.isothermal_bulk_modulus(pressure, temperature, volume, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K_S = K_T*(1. + gr * alpha * temperature)
        return K_S

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus at the pressure, temperature, and volume [Pa]
        """
        debye_T = self.__debye_temperature(params['V_0']/volume, params)
        eta_s = self.__isotropic_eta_s(params['V_0']/volume, params)

        E_th = debye.thermal_energy(temperature ,debye_T, params['n'])
        E_th_ref = debye.thermal_energy(300.,debye_T, params['n'])

        if self.order==2:
            return bm.shear_modulus_second_order(volume, params) - eta_s * (E_th-E_th_ref) / volume
        elif self.order==3:
            return bm.shear_modulus_third_order(volume, params) - eta_s * (E_th-E_th_ref) / volume
        else:
            raise NotImplementedError("")

    def heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature, and volume [J/K/mol]
        """
        debye_T = self.__debye_temperature(params['V_0']/volume, params)
        return debye.heat_capacity_v(temperature, debye_T,params['n'])

    def heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure at the pressure, temperature, and volume [J/K/mol]
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v*(1. + gr * alpha * temperature)
        return C_p

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature, and volume [1/K]
        """
        C_v = self.heat_capacity_v(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K = self.isothermal_bulk_modulus(pressure, temperature, volume, params)
        alpha = gr * C_v / K / volume
        return alpha
    

class SLB3(SLBBase):
    """
    SLB equation of state with third order finite strain expansion for the
    shear modulus (this should be preferred, as it is more thermodynamically
    consistent.
    """
    def __init__(self):
        self.order=3


class SLB2(SLBBase):
    """
    SLB equation of state with second order finite strain expansion for the
    shear modulus.  In general, this should not be used, but sometimes 
    shear modulus data is fit to a second order equation of state.  In that 
    case, you should use this.  The moral is, be careful!
    """
    def __init__(self):
        self.order=2



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import warnings

class AveragingScheme:
    """ 
    Base class defining an interface for determining average 
    elastic properties of a rock.  Given a list of volume 
    fractions for the different mineral phases in a rock, 
    as well as their bulk and shear moduli, an averaging 
    will give back a single scalar values for the averages.
    New averaging schemes should define the functions 
    average_bulk_moduli and average_shear_moduli, as
    specified here.
    """
    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite.  This defines the interface 
        for this method, and is not implemented in the base class.
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa].
 
        Returns
        -------
        
        K : float
            The average bulk modulus in [Pa].
        """
        raise NotImplementedError("")
    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite.  This defines the interface 
        for this method, and is not implemented in the base class.
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa].
 
        Returns
        -------
        
        G : float
            The average shear modulus in [Pa].
        """
        raise NotImplementedError("")
    def average_density(self, volumes, densities):
        """
        Average the densities of a composite, given a list of volume
        fractions and densitites.  This is the only method of
        :class:`burnman.averaging_schemes.averaging_scheme` which is implemented in the base class,
        as how to calculate it is not dependent on the geometry of the rock.
        The formula for density is given by

        .. math:: 
            \\rho = \\frac{\\Sigma_i \\rho_i V_i }{\\Sigma_i V_i}

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        densities : list of floats
            List of densities of each phase in the composite in [kg/m^3].
 
        Returns
        -------
        rho : float 
           Density in [kg/m^3].
        """
        total_mass = np.sum(np.array(densities)*np.array(volumes))
        total_vol = np.sum(np.array(volumes)) #should sum to one
        density = total_mass/total_vol
        return density
         
        

class VoigtReussHill(AveragingScheme):
    """ 
    Class for computing the Voigt-Reuss-Hill average for elastic properties. 
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements 
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the Voigt-Reuss-Hill average, given by:
 
        .. math::
            K_{VRH} = \\frac{K_V + K_R}{2}

        This is simply a shorthand for an arithmetic average of the bounds given 
        by :class:`burnman.averaging_schemes.voigt` and :class:`burnman.averaging_schemes.reuss`.
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
            Not used in this average.
 
        Returns
        -------
        
        K : float
            The Voigt-Reuss-Hill average bulk modulus in [Pa].
        """
        return voigt_reuss_hill_function(volumes, bulk_moduli)

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the Voigt-Reuss-Hill average, given by:
 
        .. math::
            G_{VRH} = \\frac{G_V + G_R}{2}

        This is simply a shorthand for an arithmetic average of the bounds given 
        by :class:`burnman.averaging_schemes.voigt` and :class:`burnman.averaging_schemes.reuss`.
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
            Not used in this average.
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        G : float
            The Voigt-Reuss-Hill average shear modulus in [Pa].
        """
        return voigt_reuss_hill_function(volumes, shear_moduli)


class Voigt(AveragingScheme):
    """ 
    Class for computing the Voigt (iso-strain) bound for elastic properties. 
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements 
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the Voigt (iso-strain) 
        bound, given by:
 
        .. math::
            K_V = \\Sigma_i V_i K_i 
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
            Not used in this average.
 
        Returns
        -------
        
        K : float
            The Voigt average bulk modulus in [Pa]
        """
        return voigt_average_function(volumes, bulk_moduli)

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the Voigt (iso-strain) 
        bound, given by:
 
        .. math::
            G_V = \\Sigma_i V_i G_i 
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
            Not used in this average.
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        G : float
            The Voigt average shear modulus in [Pa]
        """
        return voigt_average_function(volumes, shear_moduli)


class Reuss(AveragingScheme):
    """ 
    Class for computing the Reuss (iso-stress) bound for elastic properties. 
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements 
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the Reuss (iso-stress) 
        bound, given by:
 
        .. math::
            K_R = \\left(\\Sigma_i \\frac{V_i}{K_i} \\right)^{-1}
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
            Not used in this average.
 
        Returns
        -------
        
        K : float
            The Reuss average bulk modulus in [Pa].
        """
        return reuss_average_function(volumes, bulk_moduli)

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the Reuss (iso-stress) 
        bound, given by:
 
        .. math::
            G_R = \\left( \\Sigma_i \\frac{V_i}{G_i} \\right)^{-1}
       
        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
            Not used in this average.
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        G : float
            The Reuss average shear modulus in [Pa].
        """
        return reuss_average_function(volumes, shear_moduli)
    

class HashinShtrikmanUpper(AveragingScheme):
    """
    Class for computing the upper Hashin-Shtrikman bound for elastic properties. 
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements 
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    Implements Formulas from Watt et al (1976).  The Hashin-Shtrikman bounds
    are tighter than the Voigt and Reuss bounds because they make the 
    additional assumption that the orientation of the phases are statistically 
    isotropic.  In some cases this may be a good assumption, and in others it 
    may not be.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the upper Hashin-Shtrikman bound.
        Implements Formulas from Watt et al (1976), which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        K : float
            The upper Hashin-Shtrikman average bulk modulus in [Pa].
        """

        K_n = max(bulk_moduli)
        G_n = max(shear_moduli)

        vol_frac = volumes/sum(volumes)
 
        alpha_n = -3. / (3.*K_n+4.*G_n)
        A_n = 0
        for i in range(len(vol_frac)):
            if  bulk_moduli[i] != K_n:
                A_n = A_n + vol_frac[i]/(1./(bulk_moduli[i] - K_n) - alpha_n)

        K_upper = K_n + A_n/(1. + alpha_n*A_n)
        return K_upper

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the upper Hashin-Shtrikman bound.
        Implements Formulas from Watt et al (1976), which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        G : float
            The upper Hashin-Shtrikman average shear modulus in [Pa].
        """

        K_n = max(bulk_moduli)
        G_n = max(shear_moduli)

        vol_frac = volumes/sum(volumes)
 
        beta_n = -3. * (K_n + 2.*G_n)  / (5.*G_n * (3.*K_n+4.*G_n))
        B_n = 0
        for i in range(len(vol_frac)):
            if  shear_moduli[i] != G_n:
                B_n = B_n + vol_frac[i]/(1./(2.*(shear_moduli[i] - G_n)) - beta_n)

        G_upper = G_n + (0.5)*B_n/(1. + beta_n*B_n)
        return G_upper

class HashinShtrikmanLower(AveragingScheme):
    """
    Class for computing the lower Hashin-Shtrikman bound for elastic properties. 
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements 
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    Implements Formulas from Watt et al (1976).  The Hashin-Shtrikman bounds
    are tighter than the Voigt and Reuss bounds because they make the 
    additional assumption that the orientation of the phases are statistically 
    isotropic.  In some cases this may be a good assumption, and in others it 
    may not be.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the lower Hashin-Shtrikman bound.
        Implements Formulas from Watt et al (1976), which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        K : float
            The lower Hashin-Shtrikman average bulk modulus in [Pa].
        """

        K_1 = min(bulk_moduli)
        G_1 = min(shear_moduli)

        vol_frac = volumes/sum(volumes)
 
        alpha_1 = -3. / (3.*K_1+4.*G_1)
        A_1 = 0
        for i in range(len(vol_frac)):
            if  bulk_moduli[i] != K_1:
                A_1 = A_1 + vol_frac[i]/(1./(bulk_moduli[i] - K_1) - alpha_1)

        K_lower = K_1 + A_1/(1. + alpha_1*A_1)
        return K_lower

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the lower Hashin-Shtrikman bound.
        Implements Formulas from Watt et al (1976), which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        G : float
            The lower Hashin-Shtrikman average shear modulus in [Pa].
        """

        K_1 = min(bulk_moduli)
        G_1 = min(shear_moduli)

        vol_frac = volumes/sum(volumes)
 
        beta_1 = -3. * (K_1 + 2.*G_1)  / (5.*G_1 * (3.*K_1+4.*G_1))
        B_1 = 0
        for i in range(len(vol_frac)):
            if  shear_moduli[i] != G_1:
                B_1 = B_1 + vol_frac[i]/(1./(2.*(shear_moduli[i] - G_1)) - beta_1)

        G_lower = G_1 + (0.5)*B_1/(1. + beta_1*B_1)
        return G_lower

class HashinShtrikmanAverage(AveragingScheme):
    """ 
    Class for computing arithmetic mean of the Hashin-Shtrikman bounds on elastic properties. 
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements 
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """
    def __init__(self):
        self.upper = HashinShtrikmanUpper()
        self.lower = HashinShtrikmanLower()

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the arithmetic mean of the upper 
        and lower Hashin-Shtrikman bounds.

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
            Not used in this average.
 
        Returns
        -------
        
        K : float
            The arithmetic mean of the Hashin-Shtrikman bounds on bulk modulus in [Pa].
        """
        return (  self.upper.average_bulk_moduli(volumes, bulk_moduli, shear_moduli)\
                + self.lower.average_bulk_moduli(volumes, bulk_moduli, shear_moduli))/2.0

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the arithmetic mean of the upper 
        and lower Hashin-Shtrikman bounds.

        Parameters
        ----------
        volumes : list of floats
            List of volume fractions of each phase in the composite in [m^3].
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite in [Pa].
            Not used in this average.
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite in [Pa]. 
 
        Returns
        -------
        
        K : float
            The arithmetic mean of the Hashin-Shtrikman bounds on bulk modulus in [Pa].
        """
        return (   self.upper.average_shear_moduli(volumes, bulk_moduli, shear_moduli)\
                 + self.lower.average_shear_moduli(volumes, bulk_moduli, shear_moduli))/2.0
        
      

def voigt_average_function(phase_volume,X):
    """
    Do Voigt (iso-strain) average.  Rather like
    resistors in series.  Called by voigt and
    voigt_reuss_hill classes, takes a list of
    volumes and moduli, returns a modulus.
    """
    it = range(len(phase_volume))
    V_i = phase_volume
    V_tot = sum(V_i)
    X_voigt = sum( V_i[i]/V_tot * X[i] for i in it)
    return X_voigt

def reuss_average_function(phase_volume,X):
    """
    Do Reuss (iso-stress) average.  Rather like
    resistors in parallel.  Called by reuss and
    voigt_reuss_hill classes, takes a list of
    volumes and moduli, returns a modulus.
    """
    it = range(len(phase_volume))
    V_i = phase_volume
    V_tot = sum(V_i)
    if (min(X)<=0.0):
        X_reuss = 0.0
        warnings.warn("Oops, called reuss_average with Xi<=0!")
    else:
        X_reuss = 1./sum(  V_i[i]/V_tot* 1./X[i] for i in it)
    return X_reuss

def voigt_reuss_hill_function(phase_volume,X):
    """
    Do Voigt-Reuss-Hill average (arithmetic mean
    of Voigt and Reuss bounds).  Called by
    voigt_reuss_hill class, takes a list of
    volumes and moduli, returns a modulus.
    """
    X_vrh = (voigt_average_function(phase_volume,X) + reuss_average_function(phase_volume,X))/2.
    return X_vrh




# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
Other minerals
^^^^^^^^^^^^^^

Other minerals in burnman.minerals.other are:

"""

import burnman.mineral_helpers as bmb
from burnman.mineral import Mineral



class Speziale_fe_periclase(bmb.HelperSpinTransition):
    def __init__(self):
        bmb.HelperSpinTransition.__init__(self, 60.0e9, Speziale_fe_periclase_LS(), Speziale_fe_periclase_HS())
        self.cite = 'Speziale et al. 2007'

class Speziale_fe_periclase_HS(Mineral):
    """
    Speziale et al. 2007, Mg#=83
    """ 
    def __init__(self):
            self.params = {
                        'equation_of_state': 'mgd3',
                        'V_0': 22.9e-6,
                        'K_0': 157.5e9,
                        'Kprime_0': 3.92,
                        'molar_mass': .04567,
                        'n': 2,
                        'Debye_0': 587,
                        'grueneisen_0': 1.46,
                        'q_0': 1.2 }

class Speziale_fe_periclase_LS(Mineral):
    """
    Speziale et al. 2007, Mg#=83
    """
    def __init__(self):
        self.params = {
                        'equation_of_state': 'mgd3',
                        'V_0': 21.49e-6,
                        'K_0': 186.0e9,
                        'Kprime_0': 4.6,
                        'molar_mass': .04567,
                        'n': 2,
                        'Debye_0': 587.,
                        'grueneisen_0': 1.46,
                        'q_0': 1.2  }

    
    

        





# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import operator
import bisect
import os
import pkgutil
import numpy as np

def pretty_print_table(table,use_tabs=False):
    """
    Takes a 2d table and prints it in a nice text based format. If 
    use_tabs=True then only \t is used as a separator. This is useful for
    importing the data into other apps (Excel, ...). The default is to pad
    the columns with spaces to make them look neat. The first column is
    left aligned, while the remainder is right aligned.
    """
    if use_tabs:
        for r in table:
            print "\t".join(r).replace("_","\_")
        return        
    
    def col_width(table, colidx):
        return max([len(str(row[colidx])) for row in table])

    # create a format string with the first column left aligned, the others right  
    # example:   {:<27}{:>11}{:>6}{:>8}
    frmt = "".join([ ('{:<' if i==0 else '{:>')+str(1+col_width(table,i))+'}' for i in range(len(table[0])) ])
    for r in table:
        print frmt.format(*r)

def sort_table(table, col=0):
    """
    Sort the table according to the column number
    """
    return sorted(table, key=operator.itemgetter(col))


def float_eq(a,b):
    """
    Test if two floats are almost equal to each other
    """
    return abs(a-b)<1e-10*max(1e-5,abs(a),abs(b))


def linear_interpol(x, x1, x2, y1, y2):
    """
    Linearly interpolate to point x, between
    the points (x1,y1), (x2,y2)
    """
    assert(x1<=x)
    assert(x2>=x)
    assert(x1<=x2)

    alpha = (x - x1) / (x2-x1)
    return (1.-alpha)*y1 + alpha*y2

def read_table(filename):
    datastream = pkgutil.get_data('burnman', 'data/'+filename)
    datalines = [ line.strip() for line in datastream.split('\n') if line.strip() ]
    table=[]
    
    for line in datalines:
        if (line[0]!='#'):
            numbers = np.fromstring( line , sep =' ')
            table.append(numbers)
    return np.array(table)

def cut_table(table, min_value, max_value):
    tablen=[]
    for i in range(min_value,max_value,1):
        tablen.append(table[i,:])
    return tablen

def lookup_and_interpolate(table_x, table_y, x_value):    
    idx = bisect.bisect_left(table_x, x_value) - 1
    if (idx < 0):
        return table_y[0]
    elif (idx < len(table_x)-1):
        return linear_interpol(x_value, table_x[idx], table_x[idx+1], \
                         table_y[idx], table_y[idx+1])
    else:
        return table_y[idx]

def molar_volume_from_unit_cell_volume(unit_cell_v, z):
    """
    takes unit cell volume in Angstroms^3, as is often reported, 
    and the z number for the cell (number of atoms per unit cell,
    NOT number of atoms per molecular formula), and calculates
    the molar volume, as expected by the equations of state.
    """
    N_a = 6.0221415e23
    return  unit_cell_v*N_a/1e30/z




# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
Murakami_etal_2012
^^^^^^^^^^^^^^^^^^

Minerals from Murakami et al. (2012) supplementary table 5 and references therein, V_0 from Stixrude & Lithgow-Bertolloni 2005


"""
import burnman.mineral_helpers as bmb
from burnman.mineral import Mineral






class mg_perovskite(Mineral):
    """
    Murakami et al. (2012) supplementary table 5 and references therein, V_0 from Stixrude & Lithgow-Bertolloni 2005
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb2',
            'V_0': 24.45e-6,  # S & L-B 2005
            'K_0': 281e9,
            'Kprime_0': 4.1,
            'G_0': 173e9,
            'Gprime_0': 1.56,
            'molar_mass': .100,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.4 } 

class mg_perovskite_3rdorder(Mineral):
    """
    Murakami et al. (2012) third order fit to supplementary table 1, includes 4% Al 
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 24.45e-6,  # S & L-B 2005
            'K_0': 281e9,
            'Kprime_0': 4.1,
            'G_0': 171.42e9,
            'Gprime_0': 1.83,
            'molar_mass': .100,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.4 }

class fe_perovskite(Mineral):
    """
    Murakami et al. (2012), personal communication, Mg#=94, Al=4%
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb2',
            'V_0': 24.607e-6,
            'K_0': 251.9e9, 
            'Kprime_0': 4.01,
            'G_0': 164.7e9,
            'Gprime_0': 1.58,
            'molar_mass': .102,
            'n': 5,
            'Debye_0': 1054.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.4 } 


            
class mg_periclase(Mineral):
    """
    Murakami et al. (2012) supplementary table 5 and references therein, V_0 from Stixrude & Lithgow-Bertolloni 2005
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb2',
            'V_0': 11.24e-6,  # S & L-B 2005
            'K_0': 161e9,
            'Kprime_0': 3.9,
            'G_0': 131e9,
            'Gprime_0': 1.92,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 773., # S& L-B 2005
            'grueneisen_0': 1.5,
            'q_0': 1.5, #S&L-B 2005    
            'eta_s_0': 3.0 }             



class fe_periclase(bmb.HelperSpinTransition):
    def __init__(self):
        bmb.HelperSpinTransition.__init__(self, 63.0e9, fe_periclase_LS(), fe_periclase_HS())

class fe_periclase_3rd(bmb.HelperSpinTransition):
    def __init__(self):
        bmb.HelperSpinTransition.__init__(self, 63.0e9, fe_periclase_LS(), fe_periclase_HS())

class fe_periclase_HS(Mineral):  # From Murakami's emails, see Cayman for details, represents Mg# = .79
    """
    Murakami et al. (2012), personal communication, Mg#=79 
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb2',
            'V_0': 11.412e-6,
            'K_0': 159.1e9,
            'Kprime_0': 4.11,
            'G_0': 105.43e9,
            'Gprime_0': 1.773,
            'molar_mass': .047,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54 }

class fe_periclase_LS(Mineral):  # From Murakami's emails, see Cayman for details, represents Mg# = .79
    """
    Murakami et al. (2012), personal communication, Mg#=79
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb2',
            'V_0': 11.171e-6,
            'K_0': 170.0e9,
            'Kprime_0': 4.00,
            'G_0': 116.34e9,
            'Gprime_0': 1.668,
            'molar_mass': .047,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54}



class fe_periclase_HS_3rd(Mineral):
    """
    Murakami et al. (2012), personal communication, Mg#=92
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 11.412e-6,
            'K_0': 159.1e9,
            'Kprime_0': 4.11,
            'G_0': 129.35e9,
            'Gprime_0': 1.993,
            'molar_mass': .0469,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54 }

class fe_periclase_LS_3rd(Mineral):
    """
    Murakami et al. (2012), personal communication, Mg#=92
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 11.171e-6,
            'K_0': 170.0e9,
            'Kprime_0': 4.0,
            'G_0': 151.67e9,
            'Gprime_0': 1.754,
            'molar_mass': .0469,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54}
                



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import scipy.optimize as opt
import burnman.equation_of_state as eos

def bulk_modulus(volume, params):
    """
    compute the bulk modulus as per the third order
    birch-murnaghan equation of state.  Returns bulk
    modulus in the same units as the reference bulk
    modulus.  Pressure must be in Pa.
    """

    x = params['V_0']/volume
    f = 0.5*(pow(x, 2./3.) - 1.0)

    K = pow(1. + 2.*f, 5./2.)* (params['K_0'] + (3. * params['K_0'] * params['Kprime_0'] -  \
           5*params['K_0'] ) * f + 27./2. * (params['K_0']*params['Kprime_0'] - 4.* params['K_0'])*f*f)
    return K

def birch_murnaghan(x, params):
    """
    equation for the third order birch-murnaghan equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    """

    return 3.*params['K_0']/2. * (pow(x, 7./3.) - pow(x, 5./3.)) \
    * (1 - .75*(4-params['Kprime_0'] )*(pow(x, 2./3.) - 1))

def density(pressure, params):
    """ 
    Get the birch-murnaghan density at a reference temperature for a given
    pressure (in Pa). Returns density in kg/m^3
    """
    return params['molar_mass']/volume(pressure,params)

def volume(pressure, params):
    """
    Get the birch-murnaghan volume at a reference temperature for a given
    pressure (Pa). Returns molar volume in m^3
    """

    func = lambda x: birch_murnaghan(params['V_0']/x, params) - pressure
    V = opt.brentq(func, 0.5*params['V_0'], 1.5*params['V_0'])
    return V

def shear_modulus_second_order(volume, params):
    """
    Get the birch murnaghan shear modulus at a reference temperature, for a
    given volume.  Returns shear modulus in Pa (the same units as in
    params['G_0']).  This uses a second order finite strain expansion
    """

    x = params['V_0']/volume
    G=params['G_0'] * pow(x,5./3.)*(1.-0.5*(pow(x,2./3.)-1.)*(5.-3.*params['Gprime_0']*params['K_0']/params['G_0']))
    return G 

def shear_modulus_third_order(volume, params):
    """
    Get the birch murnaghan shear modulus at a reference temperature, for a
    given volume.  Returns shear modulus in Pa (the same units as in
    params['G_0']).  This uses a third order finite strain expansion
    """

    x = params['V_0']/volume
    f = 0.5*(pow(x, 2./3.) - 1.0)
    G = pow((1. + 2*f), 5./2.)*(params['G_0']+(3.*params['K_0']*params['Gprime_0'] - 5.*params['G_0'])*f + (6.*params['K_0']*params['Gprime_0']-24.*params['K_0']-14.*params['G_0']+9./2. * params['K_0']*params['Kprime_0'])*f*f)
    return G 


class BirchMurnaghanBase(eos.EquationOfState):
    """
    A wrapper class for the birch murnaghan functions above so that
    it satisfies the interface requirements as specified in
    equation_of_state.py
    """

    def volume(self,pressure, temperature, params):
        return volume(pressure,params)

    def isothermal_bulk_modulus(self,pressure,temperature, volume, params):
        return bulk_modulus(volume, params)

    def adiabatic_bulk_modulus(self,pressure, temperature, volume, params):
        return bulk_modulus(volume,params)

    def shear_modulus(self,pressure, temperature, volume, params):
        if(self.order == 2):
          return shear_modulus_second_order(volume,params)
        elif(self.order == 3):
          return shear_modulus_third_order(volume,params)

    # return large for heat capacity, zero for grueneisen and expansivity
    def heat_capacity_v(self,pressure, temperature, volume, params):
        return 1.e99

    def heat_capacity_p(self,pressure, temperature, volume, params):
        return 1.e99

    def thermal_expansivity(self,pressure, temperature, volume, params):
        return 0.

    def grueneisen_parameter(self,pressure,temperature,volume,params):
        return 0.


class BM3(BirchMurnaghanBase):
    def __init__(self):
        self.order=3


class BM2(BirchMurnaghanBase):
    def __init__(self):
        self.order=2



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import scipy.optimize as opt

import burnman.equation_of_state as eos
import burnman.birch_murnaghan as bm
import burnman.debye as debye


class MGDBase(eos.EquationOfState):
    """
    Base class for a generic finite-strain-mie-grueneisen-debye
    equation of state.  References for this can be found in many
    places, such as Shim, Duffym and Kenichi (2002) and Jackson and Rigedn
    (1996).  Here we mostly follow the appendices of Matas et al (2007)
    """

    #def __init__(self):
    #    pass

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume (EQ B6)
        """
        return self.__grueneisen_parameter(params['V_0']/volume, params)

    def volume(self, pressure,temperature,params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        EQ B7
        """
        func = lambda x: bm.birch_murnaghan(params['V_0']/x, params) + \
            self.__thermal_pressure(temperature, x, params) - \
            self.__thermal_pressure(300., x, params) - pressure
        V = opt.brentq(func, 0.5*params['V_0'], 1.5*params['V_0'])
        return V

    def isothermal_bulk_modulus(self, pressure,temperature,volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ B8
        """
        K_T = bm.bulk_modulus(volume, params) + \
            self.__thermal_bulk_modulus(temperature,volume, params) - \
            self.__thermal_bulk_modulus(300.,volume, params)  #EQB13
        return K_T

    #calculate the mgd shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ B11
        """
        if self.order==2:
            return bm.shear_modulus_second_order(volume,params) + \
                self.__thermal_shear_modulus(temperature,volume, params) - \
                self.__thermal_shear_modulus(300.,volume, params) # EQ B11
        elif self.order==3:
            return bm.shear_modulus_third_order(volume,params) + \
                self.__thermal_shear_modulus(temperature,volume, params) - \
                self.__thermal_shear_modulus(300.,volume, params) # EQ B11
        else:
            raise NotImplementedError("")

    #heat capacity at constant volume
    def heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature, and volume [J/K/mol]
        """
        Debye_T = self.__debye_temperature(params['V_0']/volume, params)
        C_v = debye.heat_capacity_v(temperature, Debye_T, params['n'])
        return C_v

    def thermal_expansivity(self, pressure, temperature, volume , params):
        """
        Returns thermal expansivity at the pressure, temperature, and volume [1/K]
        """
        C_v = self.heat_capacity_v(pressure,temperature,volume,params)
        gr = self.__grueneisen_parameter(params['V_0']/volume, params)
        K = self.isothermal_bulk_modulus(pressure, temperature, volume ,params)
        alpha = gr * C_v / K / volume
        return alpha

    #heat capacity at constant pressure
    def heat_capacity_p(self,pressure, temperature,volume,params):
        """
        Returns heat capacity at constant pressure at the pressure, temperature, and volume [J/K/mol]
        """
        alpha = self.thermal_expansivity(pressure,temperature,volume,params)
        gr = self.__grueneisen_parameter(params['V_0']/volume, params)
        C_v = self.heat_capacity_v(pressure,temperature,volume,params)
        C_p = C_v*(1. + gr * alpha * temperature)
        return C_p

    def adiabatic_bulk_modulus(self,pressure,temperature,volume,params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ D6
        """
        K_T= self.isothermal_bulk_modulus(pressure,temperature,volume,params)
        alpha = self.thermal_expansivity(pressure,temperature,volume,params)
        gr = self.__grueneisen_parameter(params['V_0']/volume, params)
        K_S = K_T*(1. + gr * alpha * temperature)
        return K_S

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        EQ B7
        """
        return bm.birch_murnaghan(params['V_0']/volume, params) + \
                self.__thermal_pressure(temperature,volume, params) - \
                self.__thermal_pressure(300.,volume, params)

    #calculate the thermal correction to the shear modulus as a function of V, T
    def __thermal_shear_modulus(self, T, V, params):
        gr = self.__grueneisen_parameter(params['V_0']/V, params)
        Debye_T = self.__debye_temperature(params['V_0']/V, params) 
        G_th= 3./5. * ( self.__thermal_bulk_modulus(T,V,params) - \
                 6*debye.R*T*params['n']/V * gr * debye.debye_fn(Debye_T/T) ) # EQ B10
        return G_th

    #compute the Debye temperature in K.  Takes the
    #parameter x, which is V_0/V (molar volumes).
    #Depends on the reference grueneisen parameter,
    #the reference Debye temperature, and the factor
    #q_0, see Matas eq B6
    def __debye_temperature(self, x, params):
        return params['Debye_0']*np.exp((params['grueneisen_0']- \
            self.__grueneisen_parameter(x, params))/params['q_0'])

    #compute the grueneisen parameter with depth, according
    #to q_0.  Takes x=V_0/V. See Matas eq B6
    def __grueneisen_parameter(self, x, params):
        return params['grueneisen_0']*pow(1./x, params['q_0'])

    #calculate isotropic thermal pressure, see
    # Matas et. al. (2007) eq B4
    def __thermal_pressure(self,T,V, params):
        Debye_T = self.__debye_temperature(params['V_0']/V, params) 
        gr = self.__grueneisen_parameter(params['V_0']/V, params)
        P_th = gr * debye.thermal_energy(T,Debye_T, params['n'])/V
        return P_th


    #calculate the thermal correction for the mgd
    #bulk modulus (see matas et al, 2007)
    def __thermal_bulk_modulus(self, T,V,params):
        gr = self.__grueneisen_parameter(params['V_0']/V, params)
        Debye_T = self.__debye_temperature(params['V_0']/V, params) 
        K_th = 3.*params['n']*debye.R*T/V * gr * \
            ((1. - params['q_0'] - 3.*gr)*debye.debye_fn(Debye_T/T)+3.*gr*(Debye_T/T)/(np.exp(Debye_T/T) - 1.)) # EQ B5
        return K_th


class MGD3(MGDBase):
    """
    MGD equation of state with third order finite strain expansion for the
    shear modulus (this should be preferred, as it is more thermodynamically
    consistent.
    """
    def __init__(self):
        self.order=3


class MGD2(MGDBase):
    """
    MGD equation of state with second order finite strain expansion for the
    shear modulus.  In general, this should not be used, but sometimes 
    shear modulus data is fit to a second order equation of state.  In that 
    case, you should use this.  The moral is, be careful!
    """
    def __init__(self):
        self.order=2
    



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""

SLB_2011
^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2011 and references therein 

"""

import burnman.mineral_helpers as bmb
from burnman.mineral import Mineral


class stishovite (Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2011 and references therein 
    """
    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 14.02e-6,
            'K_0': 314.0e9,
            'Kprime_0': 3.8,
            'G_0': 220.0e9,
            'Gprime_0': 1.9,
            'molar_mass': .0601,
            'n': 3,
            'Debye_0': 1108.,
            'grueneisen_0': 1.37,
            'q_0': 2.8,
            'eta_s_0': 4.6}

        self.uncertainties = {
             'err_K_0':8.e9,
             'err_Kprime_0':0.1,
             'err_G_0':12.e9,
             'err_Gprime_0':0.1,
             'err_Debye_0' : 13.,
             'err_grueneisen_0': .17,
             'err_q_0': 2.2,
             'err_eta_s_0' : 1.0 
            }


class periclase (Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2011 and references therein 
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 11.24e-6,
            'K_0': 161.0e9,
            'Kprime_0': 3.8,
            'G_0': 131.0e9,
            'Gprime_0': 2.1,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 767.,
            'grueneisen_0': 1.36,
            'q_0': 1.7, #1.7
            'eta_s_0': 2.8 } # 2.8

        self.uncertainties = {
        'err_K_0': 3.e9,
        'err_Kprime_0':.2,
        'err_G_0':1.0e9,
        'err_Gprime_0':.1,
        'err_Debye_0':9.,
        'err_grueneisen_0':.05,
        'err_q_0':.2,
        'err_eta_s_0':.2 }

class wuestite (Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2011 and references therein
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 12.26e-6,
            'K_0': 179.0e9,
            'Kprime_0': 4.9,
            'G_0': 59.0e9,
            'Gprime_0': 1.4,
            'molar_mass': .0718,
            'n': 2,
            'Debye_0': 454.,
            'grueneisen_0': 1.53,
            'q_0': 1.7, #1.7
            'eta_s_0': -0.1 } #

        self.uncertainties = {
            'err_K_0':1.e9,
            'err_Kprime_0':.2,
            'err_G_0':1.e9,
            'err_Gprime_0':.1,
            'err_Debye_0':21.,
            'err_grueneisen_0':.13,
            'err_q_0':1.0,
            'err_eta_s_0':1.0}


class ferropericlase(bmb.HelperSolidSolution):
    def __init__(self, fe_num):
        base_materials = [periclase(), wuestite()]
        molar_fraction = [1. - fe_num, 0.0 + fe_num] # keep the 0.0 +, otherwise it is an array sometimes
        bmb.HelperSolidSolution.__init__(self, base_materials, molar_fraction)

class mg_fe_perovskite(bmb.HelperSolidSolution):
    def __init__(self, fe_num):
        base_materials = [mg_perovskite(), fe_perovskite()]
        molar_fraction = [1. - fe_num, 0.0 + fe_num] # keep the 0.0 +, otherwise it is an array sometimes
        bmb.HelperSolidSolution.__init__(self, base_materials, molar_fraction)

class mg_perovskite(Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2011 and references therein  
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 24.45e-6,
            'K_0': 251.0e9,   
            'Kprime_0': 4.1,     
            'G_0': 173.0e9,  
            'Gprime_0': 1.7,  
            'molar_mass': .1000,
            'n': 5,
            'Debye_0': 905.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'eta_s_0': 2.6 } #2.6

        self.uncertainties = {
            'err_K_0': 3.e9,
            'err_Kprime_0': 0.1,
            'err_G_0': 2.e9,
            'err_Gprime_0' : 0.0,
            'err_Debye_0': 5.,
            'err_grueneisen_0':.05,
            'err_q_0': .3,
            'err_eta_s_0':.3}


class fe_perovskite(Mineral):
    """
    Stixrude & Lithgow-Bertelloni 2011 and references therein
    """
    def __init__(self):
        self.params = {
            'equation_of_state':'slb3',
            'V_0': 25.49e-6,
            'K_0': 272.0e9, 
            'Kprime_0': 4.1,  
            'G_0': 133.0e9,
            'Gprime_0': 1.4,   
            'molar_mass': .1319, 
            'n': 5,
            'Debye_0': 871.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'eta_s_0': 2.3 } #2.3

        self.uncertainties = {
            'err_K_0':40e9,
            'err_Kprime_0':1.,
            'err_G_0':40e9,
            'err_Gprime_0':0.0,
            'err_Debye_0':26.,
            'err_grueneisen_0':.3,
            'err_q_0':1.0,
            'err_eta_s_0':1.0}

class mg_fe_perovskite_pt_dependent(bmb.HelperFeDependent):
    def __init__(self, iron_number_with_pt, idx):
        bmb.HelperFeDependent.__init__(self, iron_number_with_pt, idx)

    def create_inner_material(self, iron_number):
        return mg_fe_perovskite(iron_number)

class ferropericlase_pt_dependent(bmb.HelperFeDependent):
    def __init__(self, iron_number_with_pt, idx):
        bmb.HelperFeDependent.__init__(self, iron_number_with_pt, idx)

    def create_inner_material(self, iron_number):
        return ferropericlase(iron_number)



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import scipy.integrate as integrate

import burnman.geotherm
import burnman.seismic
import burnman.averaging_schemes

class ElasticProperties:
    """
    Class that contains volume, density, and moduli. This is generated for
    different pressures and temperatures using
    :func:`calculate_moduli`. Several instances of elastic_properties can be
    averaged using :func:`average_moduli`.

    :var float V: volume in [m^3] 
    :var float rho: density in [kg/m^3]
    :var float K: bulk modulus K in [Pa] 
    :var float G: shear modulus G in [Pa] 
    """

    def __init__(self, V=None, rho=None, K=None, G=None, fraction=None):
        """
        create an object with the given parameters.
        """
        self.V = V
        self.rho = rho
        self.K = K
        self.G = G
        self.fraction = fraction


def calculate_moduli(rock, pressures, temperatures):
    """
    Given a composite and a list of pressures [Pa] and temperatures [K],
    calculate the elastic moduli and densities of the individual phases.

    :param burnman.abstract_material rock: this is a rock

    :type pressures: list of float
    :param pressures: list of pressures you want to evaluate the rock at. In
                      [Pa].

    :type temperatures: list of float
    :param temperatures: list of temperatures you want to evaluate the rock at. In [K].

    :returns: 
      answer -- an array of (n_evaluation_points by n_phases) of
      elastic_properties(), so the result is of the form
      answer[pressure_idx][phase_idx].V
    :rtype: list of list of :class:`burnman.elastic_properties`
    """

    answer = [[] for p in pressures]

    for idx in range(len(pressures)):
        rock.set_state(pressures[idx], temperatures[idx])
        (fractions,minerals) = rock.unroll()
        for (fraction,mineral) in zip(fractions,minerals):
            e = ElasticProperties()
            e.V = fraction * mineral.molar_volume()
            e.K = mineral.adiabatic_bulk_modulus()
            e.G = mineral.shear_modulus()
            e.rho = mineral.molar_mass() / mineral.molar_volume()
            e.fraction = fraction
            answer[idx].append(e)
        
    return answer

def average_moduli(moduli_list, averaging_scheme=burnman.averaging_schemes.VoigtReussHill):
    """
    Given an array of (n_evaluation_points by n_phases) of
    :class:`elastic_properties` (as, for instance, generated by
    :func:`calculate_moduli`), calculate the bulk properties, according to
    some averaging scheme. The averaging scheme defaults to Voigt-Reuss-Hill
    (see :class:`burnman.averaging_schemes.voigt_reuss_hill`), but the user
    may specify, Voigt, Reuss, the Hashin-Shtrikman bounds, or any user
    defined scheme that satisfies the interface
    :class:`burnman.averaging_schemes.averaging_scheme` (also see
    :doc:`averaging`).

    :type moduli_list: list of list of :class:`burnman.elastic_properties`
    :param moduli_list: bla

    :type averaging_scheme: :class:`burnman.averaging_schemes.averaging_scheme`
    :param averaging_scheme: Averaging scheme to use.
    
    :returns: A list of n_evaluation_points instances of elastic_properties.
    :rtype: list of :class:`burnman.elastic_properties`
    """
    n_pressures = len(moduli_list)
    result = [ElasticProperties() for i in range(n_pressures)]
    
    for idx in range(n_pressures):
        fractions = np.array([e.fraction for e in moduli_list[idx]])

        V_frac = np.array([m.V for m in moduli_list[idx]])
        K_ph = np.array([m.K for m in moduli_list[idx]])
        G_ph = np.array([m.G for m in moduli_list[idx]])
        rho_ph = np.array([m.rho for m in moduli_list[idx]])
               
        result[idx].V = sum(V_frac)
        result[idx].K = averaging_scheme.average_bulk_moduli(V_frac, K_ph, G_ph)
        result[idx].G = averaging_scheme.average_shear_moduli(V_frac, K_ph, G_ph)
        result[idx].rho = averaging_scheme.average_density(V_frac, rho_ph)
        result[idx].fraction = 1.0

    return result

def compute_velocities(moduli):
    """
    Given a list of elastic_properties, compute the seismic velocities Vp, Vs,
    and Vphi for each entry in the list.

    
    :type moduli: list of :class:`ElasticProperties`
    :param moduli: input elastic properties.

    :returns: lists of Vp, Vs, Vphi (in m/s each)
    :rtype: list of float, list of float, list of float


    """
    mat_vs = np.ndarray(len(moduli))
    mat_vp = np.ndarray(len(moduli))
    mat_vphi = np.ndarray(len(moduli))
    
    for i in range(len(moduli)):

        mat_vs[i] = np.sqrt( moduli[i].G / moduli[i].rho)
        mat_vp[i] = np.sqrt( (moduli[i].K + 4./3.*moduli[i].G) / moduli[i].rho)
        mat_vphi[i] = np.sqrt( moduli[i].K / moduli[i].rho)
    
    return mat_vp, mat_vs, mat_vphi
 
 
def velocities_from_rock(rock, pressures, temperatures, averaging_scheme=burnman.averaging_schemes.VoigtReussHill()):
    """
    A function that rolls several steps into one: given a rock and a list of
    pressures and temperatures, it calculates the elastic moduli of the
    individual phases using calculate_moduli(), averages them using
    average_moduli(), and calculates the seismic velocities using
    compute_velocities().


    :param burnman.abstract_material rock: this is a rock

    :type pressures: list of float
    :param pressures: list of pressures you want to evaluate the rock at. In [Pa].

    :type temperatures: list of float
    :param temperatures: list of temperatures you want to evaluate the rock at. In[K].


    :type averaging_scheme: :class:`burnman.averaging_schemes.averaging_scheme`
    :param averaging_scheme: Averaging scheme to use.

    :returns: density[kg/m^3], Vp[m/s],Vs[m/s],Vphi[m/s], bulk modulus K[Pa],shear modulus G[Pa]
    :rtype: lists of floats

    """
    moduli_list = calculate_moduli(rock, pressures, temperatures)
    moduli = average_moduli(moduli_list, averaging_scheme)
    mat_vp, mat_vs, mat_vphi = compute_velocities(moduli)
    mat_rho = np.array([m.rho for m in moduli])
    mat_K = np.array([m.K for m in moduli])
    mat_G = np.array([m.G for m in moduli])
    return mat_rho, mat_vp, mat_vs, mat_vphi, mat_K, mat_G

def depths_for_rock(rock,pressures, temperatures,averaging_scheme=burnman.averaging_schemes.VoigtReussHill()):
    """
    Function computes the self-consistent depths (to avoid using the PREM depth-pressure conversion) (Cammarano, 2013).
    It is simplified by taking g from PREM.
        
    :param burnman.abstract_material rock: this is a rock
        
    :type pressures: list of float
    :param pressures: list of pressures you want to evaluate the rock at. In [Pa].
        
    :type temperatures: list of float
    :param temperatures: list of temperatures you want to evaluate the rock at. In [K].
        
    :type averaging_scheme: :class:`burnman.averaging_schemes.averaging_scheme`
    :param averaging_scheme: Averaging scheme to use.
        
    :returns: depth [m]
    :rtype: list of floats
    """
    moduli_list = calculate_moduli(rock, pressures, temperatures)
    moduli = average_moduli(moduli_list, averaging_scheme)
    mat_rho = np.array([m.rho for m in moduli])
    seismic_model = burnman.seismic.PREM()
    depthsref = np.array(map(seismic_model.depth,pressures))
    pressref = np.zeros_like(pressures)
    g  = seismic_model.grav(depthsref) # G for prem
    depths  = np.hstack((depthsref[0],depthsref[0]+integrate.cumtrapz(1./(g*mat_rho),pressures)))
    return depths

def pressures_for_rock(rock, depths, T0, averaging_scheme=burnman.averaging_schemes.VoigtReussHill()):
    """
    Function computes the self-consistent pressures (to avoid using the PREM depth pressure conversion) (Cammarano, 2013).
    Only simplification is using g from PREM.'
    
    :param burnman.abstract_material rock: this is a rock
    
    :type depths: list of float
    :param depths: list of depths you want to evaluate the rock at. In [m].
    
    :type temperatures: list of float
    :param temperatures: list of temperatures you want to evaluate the rock at. In[K].
    
    :type averaging_scheme: :class:`burnman.averaging_schemes.averaging_scheme`
    :param averaging_scheme: Averaging scheme to use.
    
    :returns: pressures [Pa]
    :rtype: list of floats
    
    """
    # use PREM pressures as inital guestimate
    seismic_model = burnman.seismic.PREM()
    pressures,_,_,_,_ = seismic_model.evaluate_all_at(depths)
    pressref = np.zeros_like(pressures)
    #gets table with PREM gravities
    g = seismic_model.grav(depths)
    #optimize pressures for this composition
    while nrmse(len(pressures),pressures,pressref)>1.e-6:
        # calculate density
        temperatures = burnman.geotherm.adiabatic(pressures,T0,rock)
        moduli_list = calculate_moduli(rock, pressures, temperatures)
        moduli = average_moduli(moduli_list, averaging_scheme)
        mat_rho = np.array([m.rho for m in moduli])
        # calculate pressures
        pressref = pressures
        pressures = np.hstack((pressref[0], pressref[0]+integrate.cumtrapz(g*mat_rho,depths)))
    return pressures

def apply_attenuation_correction(v_p,v_s,v_phi,Qs,Qphi):
    """
    Returns lists of corrected velocities  for a given Qs (/Qmu) and Qphi
    

    
    :type Vp: list of float
    :param Vp: list of P wave velocties. In [m/s].
    :type Vs: list of float
    :param Vs: list of P wave velocties. In [m/s].
    :type Vphi: list of float
    :param Vphi: list of P wave velocties. In [m/s].
    
    :returns: Vp[m/s],Vs[m/s],Vphi[m/s]
    :rtype: list of floats
    
    """
    
    length = len(v_p)
    ret_v_p = np.zeros(length)
    ret_v_s = np.zeros(length)
    ret_v_phi = np.zeros(length)
    for i in range(length):
        ret_v_p[i],ret_v_s[i],ret_v_phi[i] = \
            burnman.seismic.attenuation_correction(v_p[i], v_s[i], v_phi[i],Qs,Qphi)
    
    return ret_v_p, ret_v_s, ret_v_phi


def compare_l2(depth,calc, obs):
    """
        
    Computes the L2 norm for N profiles at a time (assumed to be linear between points).

    .. math:: math does not work yet...
       \sum_{i=1}^{\\infty} x_{i}

    :type depths: array of float
    :param depths: depths in m.
    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]
    
    :returns: array of L2 norms of length N
    :rtype: array of floats
    """
    err=[]
    for l in range(len(calc)):
        err.append(l2(depth,calc[l],obs[l]))

    return err

def compare_chifactor(calc, obs):
    """
        
    Computes the chi factor for N profiles at a time. Assumes a 1% a priori uncertainty on the seismic model.
        

    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]
        
    :returns: error array of length N
    :rtype: array of floats
    """
    err=[]
    for l in range(len(calc)):
        err.append(chi_factor(calc[l],obs[l]))
    
    return err

def l2(x,funca,funcb):
    """
        
    Computes the L2 norm for one profile(assumed to be linear between points).
        
    :type x: array of float
    :param x: depths in m.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to
        
    :returns: L2 norm
    :rtype: array of floats
    """
    diff=np.array(funca-funcb)
    diff=diff*diff
    return integrate.trapz(diff,x)


def nrmse(x,funca,funcb):
    """ 
    Normalized root mean square error for one profile
    :type x: array of float
    :param x: depths in m.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to
    
    :returns: RMS error
    :rtype: array of floats
    
    """
    diff=np.array(funca-funcb)
    diff=diff*diff
    rmse=np.sqrt(np.sum(diff)/x)
    nrmse=rmse/(np.max(funca)-np.min(funca))
    return nrmse

def chi_factor(calc,obs):
    """
    Chi factor for one profile assuming 1% uncertainty on the reference model (obs)
    :type calc: list of arrays of float
    :param calc: array calculated values
    :type obs: list of arrays of float
    :param obs: array of reference values to compare to
        
    :returns: chi factor
    :rtype: array of floats
        
    """

    err=np.empty_like(calc)
    for i in range(len(calc)):
        err[i]=pow((calc[i]-obs[i])/(0.01*np.mean(obs)),2.)

    err_tot=np.sum(err)/len(err)

    return err_tot



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.


class Material:
    """
    Base class for all materials. The main functionality is unroll() which
    returns a list of objects of type burnman.mineral and their molar
    fractions.

    The user needs to call set_method() once in the beginning and set_state()
    before querying the material with unroll() or density().
    """

    def set_method(self, method):
        """
        """
        raise NotImplementedError("need to implement set_method() in derived class!")

    def to_string(self):
        """
        return the name of the composite
        """
        return "'" + self.__class__.__name__ + "'"

    def debug_print(self):
        print self.to_string()

    def set_state(self, pressure, temperature):
        """
        Set the material to the given pressure and temperature
        """
        self.pressure = pressure
        self.temperature = temperature

    def unroll(self):
        """ return (fractions, minerals) where both are arrays. May depend on current state """
        raise NotImplementedError("need to implement unroll() in derived class!")
        return ()

    def density(self):
        raise NotImplementedError("need to implement density() in derived class!")
        return inf            



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
This module provides several helper minerals/materials.

"""


import numpy as np
import warnings

from burnman.material import Material

from burnman.mineral import Mineral
import burnman.equation_of_state as eos
import burnman.birch_murnaghan as bm
import burnman.slb as slb
import burnman.mie_grueneisen_debye as mgd


class HelperSolidSolution(Mineral):
    """
    Class for coming up with a new mineral based based on a solid
    solution between two or more end member minerals.  It is not
    completely clear how to do this, or how valid this approximation
    is, but here we just do a weighted arithmetic average of the 
    thermoelastic properties of the end members according to their molar fractions
    """
    def __init__(self, base_materials, molar_fraction):
        """
        Takes a list of end member minerals, and a matching list of
        molar fractions of those minerals for mixing them.  Simply
        comes up with a new mineral by doing a weighted arithmetic
        average of the end member minerals
        """
        self.base_materials = base_materials
        self.molar_fraction = molar_fraction
        assert(len(base_materials) == len(molar_fraction))
        assert(sum(molar_fraction) > 0.9999)
        assert(sum(molar_fraction) < 1.0001)

        #does not make sense to do a solid solution with different number of 
        #atoms per formula unit, at least not simply...
        for m in base_materials:
            if(base_materials[0].params.has_key('n')):
                assert(m.params['n'] == base_materials[0].params['n'])

    def set_state(self, pressure, temperature):
        for mat in self.base_materials:
            mat.method = self.method
            mat.set_state(pressure, temperature)

        itrange = range(0, len(self.base_materials))

        self.params = {}

        # some do arithmetic averaging of the end members
        for prop in self.base_materials[0].params:
           try:
               self.params[prop] = sum([ self.base_materials[i].params[prop] * self.molar_fraction[i] for i in itrange ])
           except TypeError:
               #if there is a type error, it is probably a string.  Just go with the value of the first base_material.
               self.params[prop] = self.base_materials[0].params[prop]
        Mineral.set_state(self, pressure, temperature)

class HelperSpinTransition(Material):
    """ 
    Helper class that makes a mineral that switches between two materials
    (for low and high spin) based on some transition pressure [Pa]
    """
    
    def __init__(self, transition_pressure, ls_mat, hs_mat):
        """ 
        Takes a transition pressure, and two minerals.  Use the 
        thermoelastic parameters for ls_mat below the transition
        pressure, and the thermoelastic parameters for hs_mat 
        above the transition pressure
        """
        self.transition_pressure = transition_pressure
        self.ls_mat = ls_mat
        self.hs_mat = hs_mat
        self.active_mat = None
                
    def set_method(self, method):
        self.ls_mat.set_method(method)
        self.hs_mat.set_method(method)

    def set_state(self, pressure, temperature):
        if (pressure >= self.transition_pressure):
            self.active_mat = self.ls_mat
        else:
            self.active_mat = self.hs_mat
            
        self.active_mat.set_state(pressure, temperature)

    def unroll(self):
        """ return (fractions, minerals) where both are arrays. May depend on current state """
        return ([1.0],[self.active_mat])

    def density(self):
        return self.active_mat.density()






class HelperFeDependent(Material):

    """
    Helper to implement a rock that does iron exchange (two minerals with
    changing iron content based on p,T).

    Classes deriving from this helper need to implement
    create_inner_material() and provide a function in __init__ that computes
    the iron exchange.
    """

    def __init__(self, iron_number_with_pt, idx):
        self.iron_number_with_pt = iron_number_with_pt
        self.which_index = idx  # take input 0 or 1 from iron_number_with_pt()

    def create_inner_material(self, iron_number):
        raise NotImplementedError("need to implement create_inner_material() in derived class!")
        return None

    def set_method(self, method):
        self.method = method

    def iron_number(self):
        return self.iron_number_with_pt(self.pressure, self.temperature)[self.which_index]

    def set_state(self, pressure, temperature):
        Material.set_state(self, pressure, temperature)
        self.base_material = self.create_inner_material(self.iron_number())
        self.base_material.set_method(self.method)
        self.base_material.set_state(pressure, temperature)

    def unroll(self):
        """ return (fractions, minerals) where both are arrays. May depend on current state """
        return ([1.0],[self.base_material])

    def density(self):
        return self.base_material.density()





# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import warnings

from burnman.material import Material
import burnman.slb as slb
import numpy as np
import burnman.equation_of_state as eos
import burnman.birch_murnaghan as bm
import burnman.slb as slb
import burnman.mie_grueneisen_debye as mgd

class Mineral(Material):
    """
    This is the base class for all minerals. States of the mineral
    can only be queried after setting the pressure and temperature
    using set_state(). The method for computing properties of
    the material is set using set_method(), which should be done
    once after creating the material.
    
    If deriving from this class, set the properties in self.params
    to the desired values. For more complicated materials you
    can overwrite set_state(), change the params and then call
    set_state() from this class.

    All the material parameters are expected to be in plain SI units.  This
    means that the elastic moduli should be in Pascals and NOT Gigapascals,
    and the Debye temperature should be in K not C.  Additionally, the
    reference volume should be in m^3/(mol molecule) and not in unit cell
    volume and 'n' should be the number of atoms per molecule.  Frequently in
    the literature the reference volume is given in Angstrom^3 per unit cell.
    To convert this to m^3/(mol molecule) you should multiply by 10^(-30) *
    N_a / Z, where N_a is Avogadro's number and Z is the number of atoms per
    unit cell.  You can look up Z in many places, including www.mindat.org
    """

    def __init__(self):
        self.params = {    'name':'generic',
            'equation_of_state': 'slb3', #Equation of state used to fit the parameters
            'V_0': 0., #Molar volume [m^3/(mole molecules)] at room pressure/temperature
            'K_0': 0., #Reference bulk modulus [Pa] at room pressure/temperature
            'Kprime_0': 0., #pressure derivative of bulk modulus
            'G_0': 0., #reference shear modulus at room pressure/temperature
            'Gprime_0': 0., #pressure derivative of shear modulus
            'molar_mass': 0., #molar mass in units of [kg/mol]
            'n': 0., #number of atoms per molecule
            'Debye_0': 0., #Debye temperature for material. See Stixrude & Lithgow-Bertelloni, 2005 for values 
            'grueneisen_0': 0., #Gruneisen parameter for material. See Stixrude & Lithgow-Bertelloni, 2005 for values
            'q_0': 0., #q value used in caluclations. See Stixrude & Lithgow-Bertelloni, 2005 for values
            'eta_s_0': 0.0} #eta value used in calculations. See Stixrude & Lithgow-Bertelloni, 2005 for values
        self.method = None

    def set_method(self, method):
        """
        Set the equation of state to be used for this mineral.
        Takes a string corresponding to any of the predefined
        equations of state:  'bm2', 'bm3', 'mgd2', 'mgd3', 'slb2',
        or 'slb3'.  Alternatively, you can pass a user defined
        class which derives from the equation_of_state base class.
        """
        if( isinstance(method, basestring)):
            if (method == "slb2"):
                self.method = slb.SLB2()
            elif (method == "mgd2"):
                self.method = mgd.MGD2()
            elif (method == "mgd3"):
                self.method = mgd.MGD3()
            elif (method == "slb3"):
                self.method = slb.SLB3()
            elif (method == "bm2"):
                self.method = bm.BM2()
            elif (method == "bm3"):
                self.method = bm.BM3()
            else:
                raise Exception("unsupported material method " + method)
        elif ( issubclass(method, eos.EquationOfState) ):
            self.method = method()
        else:
            raise Exception("unsupported material method " + method.__class__.__name__ )

    def to_string(self):
        """
        Returns the name of the mineral class
        """
        return "'" + self.__class__.__module__.replace(".minlib_",".") + "." + self.__class__.__name__ + "'"

    def unroll(self):
        return ([1.0],[self])

    def set_state(self, pressure, temperature):
        """
        Update the material to the given pressure [Pa] and temperature [K].
        
        This updates the other properties of this class (v_s, v_p, ...).
        """

        #in an effort to avoid additional work, don't do all the calculations if nothing has changed
        try:
            if self.pressure == pressure and self.temperature == temperature and self.old_params == self.params:
                return
        except AttributeError:
            pass  #do nothing

        self.pressure = pressure
        self.temperature = temperature
        self.old_params = self.params
        
        self.V = self.method.volume(self.pressure, self.temperature, self.params)
        self.gr = self.method.grueneisen_parameter(self.pressure, self.temperature, self.V, self.params)
        self.K_T = self.method.isothermal_bulk_modulus(self.pressure, self.temperature, self.V, self.params)
        self.K_S = self.method.adiabatic_bulk_modulus(self.pressure, self.temperature, self.V, self.params)
        self.C_v = self.method.heat_capacity_v(self.pressure, self.temperature, self.V, self.params)
        self.C_p = self.method.heat_capacity_p(self.pressure, self.temperature, self.V, self.params)
        self.alpha = self.method.thermal_expansivity(self.pressure, self.temperature, self.V, self.params)
        
        if (self.params.has_key('G_0') and self.params.has_key('Gprime_0')):
            self.G = self.method.shear_modulus(self.pressure, self.temperature, self.V, self.params)
        else:    
            self.G = float('nan') #nan if there is no G, this should propagate through calculations to the end
            warnings.warn(('Warning: G_0 and or Gprime_0 are undefined for ' + self.to_string()))

    def molar_mass(self):
        """
        Returns molar mass of the mineral [kg/mol]
        """
        return self.params['molar_mass']

    def density(self):
        """
        Returns density of the mineral [kg/m^3]
        """
        return  self.params['molar_mass'] / self.V

    def molar_volume(self):
        """
        Returns molar volume of the mineral [m^3/mol]
        """
        return self.V
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the mineral [unitless]
        """
        return self.gr
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the mineral [Pa]
        """
        return self.K_T
    def adiabatic_bulk_modulus(self):
        """
        Returns adiabatic bulk modulus of the mineral [Pa]
        """
        return self.K_S
    def shear_modulus(self):
        """
        Returns shear modulus of the mineral [Pa]
        """
        return self.G
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient of the mineral [1/K]
        """
        return self.alpha
    def heat_capacity_v(self):
        """
        Returns heat capacity at constant volume of the mineral [J/K/mol]
        """
        return self.C_v
    def heat_capacity_p(self):
        """
        Returns heat capacity at constant pressure of the mineral [J/K/mol]
        """
        return self.C_p
    def v_s(self):
        """
        Returns shear wave speed of the mineral [m/s]
        """
        return np.sqrt(self.shear_modulus() / \
            self.density()) 
    def v_p(self):
        """
        Returns P wave speed of the mineral [m/s]
        """
        return np.sqrt((self.adiabatic_bulk_modulus() + 4. / 3. * \
            self.shear_modulus()) / self.density()) 
    def v_phi(self):
        """
        Returns bulk sound speed of the mineral [m/s]
        """
        return np.sqrt(self.adiabatic_bulk_modulus() / self.density())



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
BurnMan
=======

BurnMan is an open source mineral physics toolbox written in Python to
determine seismic velocities for the lower mantle. BurnMan calculates the
isotropic thermoelastic moduli by solving the equations-of-state for a
mixture of minerals defined by the user. The user may select from a list of
minerals applicable to the lower mantle included or easily define one of
their own.

Features:
  - form composites of arbitrary combination of :doc:`minerals`
  - extensive :doc:`mineral_database`
  - easy plotting and comparison of seismic profiles using matplotlib
  - many examples highlighting different features of BurnMan
  - different thermoelastic models, choice between second or third order accuracy
  - different averaging schemes
  - different geotherms
  - extensible: all parts can be replaced by user-written modules if desired



"""


#classes for representing rocks and minerals
from mineral import Mineral
from material import Material
from composite import Composite
from mineral_helpers import *

#high level functions
from main import *

#mineral library
import minerals

#central user tools
import seismic
import averaging_schemes
import geotherm

#miscellaneous
import tools
from partitioning import calculate_partition_coefficient,calculate_phase_percents



from setuptools import setup, find_packages

setup ( name = 'burnman',
        version = '0.6b3',
        description='a lower mantle toolkit',
        url='http://burnman.org',
        author='Ian Rose',
        author_email='ian.rose@berkeley.edu',
        license='GPL',
        long_description='BurnMan is a lower mantle shear velocity generator constrained by mineral physics.',
        packages = ['burnman', 'burnman.minerals'],
        package_data = { 'burnman' : ['data/input_*/*'] },
        install_requires=['numpy', 'matplotlib', 'scipy']
       )



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import scipy.integrate as integrate
from numpy.polynomial.chebyshev import Chebyshev

"""
Functions for the Debye model.  Note that this is not Mie-Grueneisen-Debye, 
just Debye, so is pretty limited.  Combine this with Mie-Grueneisen and
Birch-Murnaghan to get a full EOS
"""

R = 8.314462175



def debye_fn(x):
    """
    Evaluate the Debye function.  Takes the parameter
    xi = Debye_T/T
    """
    sol = integrate.quad( lambda xi: pow(xi,3.)/(np.exp(xi)-1.) , 0.0, x) # EQ B3
    return 3.*sol[0]/pow(x,3.)



chebyshev_representation = Chebyshev( [ 2.707737068327440945/2.0, 0.340068135211091751, -0.12945150184440869e-01, \
                                     0.7963755380173816e-03, -0.546360009590824e-04, 0.39243019598805e-05, \
                                    -0.2894032823539e-06, 0.217317613962e-07, -0.16542099950e-08, \
                                     0.1272796189e-09, -0.987963460e-11, 0.7725074e-12, -0.607797e-13, \
                                     0.48076e-14, -0.3820e-15, 0.305e-16, -0.24e-17] )
eps = np.finfo(np.float).eps
sqrt_eps = np.sqrt(np.finfo(np.float).eps)
log_eps = np.log(np.finfo(np.float).eps)

def debye_fn_cheb(x):
    """
    Evaluate the Debye function using a Chebyshev series expansion coupled with 
    asymptotic solutions of the function.  Shamelessly adapted from the GSL implementation
    of the same function (Itself adapted from Collected Algorithms from ACM).  
    Should give the same result as debye_fn(x) to near machine-precision.
    """
    val_infinity = 19.4818182068004875;
    xcut = -log_eps

    assert(x > 0.0) #check for invalid x

    if x < 2.0*np.sqrt(2.0)*sqrt_eps:
        return 1.0 - 3.0*x/8.0 + x*x/20.0;
    elif x <= 4.0 :
        t = x*x/8.0 - 1.0;
        c = chebyshev_representation(t)
        return c - 0.375*x;
    elif x < -(np.log(2.0) + log_eps ):
        nexp = int(np.floor(xcut/x));
        ex  = np.exp(-x);
        xk  = nexp * x;
        rk  = nexp;
        sum = 0.0;
        for i in range(nexp,0,-1):
            xk_inv = 1.0/xk;
            sum *= ex;
            sum += (((6.0*xk_inv + 6.0)*xk_inv + 3.0)*xk_inv + 1.0) / rk;
            rk -= 1.0;
            xk -= x;
        return val_infinity/(x*x*x) - 3.0 * sum * ex;
    elif x < xcut:
        x3 = x*x*x;
        sum = 6.0 + 6.0*x + 3.0*x*x + x3;
        return  (val_infinity - 3.0 * sum * np.exp(-x)) / x3;
    else:
        return ((val_infinity/x)/x)/x;


def thermal_energy(T, debye_T, n):
    """
    calculate the thermal energy of a substance.  Takes the temperature,
    the Debye temperature, and n, the number of atoms per molecule.
    Returns thermal energy in J/mol
    """
    if T == 0:
        return 0
    E_th = 3.*n*R*T * debye_fn_cheb(debye_T/T)
    return E_th

def heat_capacity_v(T,debye_T,n):
    """
    Heat capacity at constant volume.  In J/K/mol
    """
    if T ==0:
        return 0
    x = debye_T/T
    C_v = 3.0*n*R* ( 4.0*debye_fn_cheb(x) - 3.0*x/(np.exp(x)-1.0) )
    return C_v




if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import time

    def old_thermal(T, debye_T, n):
        if T == 0:
            return 0
        return 3.*n*R*T * debye_fn(debye_T/T)

    def old_heat(T, debye_T, n):
        if T ==0:
            return 0
        deb = integrate.quad( lambda x : pow(x,4.)*np.exp(x)/pow((np.exp(x)-1.),2.), 0.0, debye_T/T)
        return 9.*n*R*deb[0]/pow(debye_T/T,3.)

    temperatures = np.linspace(100,5000, 10000)
    Debye_T = 1000.
    old = np.empty_like(temperatures)
    start = time.clock()
    for i in range(len(temperatures)):
        old[i] = old_heat(temperatures[i], Debye_T, 1.0)
    time_old = time.clock()-start

    new = np.empty_like(temperatures)
    start = time.clock()
    for i in range(len(temperatures)):
        new[i] = heat_capacity_v(temperatures[i], Debye_T, 1.0)
    time_new = time.clock()-start
    
    print "error %e"%np.linalg.norm((old-new)/new)
    print "time old %g, time new %g"%(time_old,time_new)



    temperatures = np.linspace(0,5000, 200)
    vibrational_energy = np.empty_like(temperatures)
    heat_capacity = np.empty_like(temperatures)
    Debye_T = 1000.
    for i in range(len(temperatures)):
      vibrational_energy[i] = thermal_energy(temperatures[i], Debye_T, 1.0)
      heat_capacity[i] = heat_capacity_v(temperatures[i], Debye_T, 1.0)

    plt.subplot(121)
    plt.plot(temperatures, vibrational_energy)
    plt.subplot(122)
    plt.plot(temperatures, heat_capacity)
    plt.show()


    



# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

import numpy as np
import matplotlib.pyplot as pyplot
import scipy.integrate as integrate

import burnman.tools
import burnman.seismic


def brown_shankland(pressure):
    """
    Geotherm from Brown and Shankland (1981).

    Parameters
    ----------
    pressure : list of floats
        The list of pressures in [Pa] at which to evaluate the geotherm.

    Returns
    -------
    temperature : list of floats
        The list of temperatures in [K] for each of the pressures
    """
    temperature = np.empty_like(pressure)
    for i in range(len(pressure)):
      depth = burnman.seismic.prem_model.depth(pressure[i])
      temperature[i] = burnman.tools.lookup_and_interpolate(table_brown_depth, table_brown_temperature, depth)    
    return temperature


def anderson(pressure):
    """
    Geotherm from Anderson (1982).

    Parameters
    ----------
    pressure : list of floats
        The list of pressures in [Pa] at which to evaluate the geotherm.

    Returns
    -------
    temperature : list of floats
        The list of temperatures in [K] for each of the pressures
    """
    temperature = np.empty_like(pressure)
    for i in range(len(pressure)):
      depth = burnman.seismic.prem_model.depth(pressure[i])
      temperature[i] = burnman.tools.lookup_and_interpolate(table_anderson_depth, table_anderson_temperature, depth)    
    return temperature

def adiabatic(pressures, T0, rock):
    """
    This calculates a geotherm based on an anchor temperature and a rock, 
    assuming that the rock's temperature follows an adiabatic gradient with 
    pressure.  This amounts to integrating 

    .. math:: 
        \\frac{\partial T}{\partial P} = \\frac{ \\gamma  T}{ K_s } 

    where :math:`\\gamma` is the Grueneisen parameter and :math:`K_s` is 
    the adiabatic bulk modulus.  
 
    Parameters
    ----------

    pressures : list of floats
        The list of pressures in [Pa] at which to evaluate the geotherm.
   
    T0 : float
        An anchor temperature, corresponding to the temperature of the first
        pressure in the list.

    rock : :class:`burnman.composite`
        Material for which we compute the adiabat.  From this material we 
        must compute average Grueneisen parameters and adiabatic bulk moduli
        for each pressure/temperature. 

    Returns
    -------

    temperature: list of floats
        The list of temperatures in [K] for each of the pressures
    """
    temperatures = integrate.odeint(lambda t,p : dTdP(t,p,rock), T0, pressures)
    return temperatures.ravel()

def dTdP(temperature, pressure, rock):
    """
    ODE to integrate temperature with depth for a composite material
    Assumes that the minerals exist at a common pressure (Reuss bound, should be good for
    slow deformations at high temperature), as well as an adiabatic process.  This
    corresponds to conservation of enthalpy.  
    First consider compression of the composite to a new pressure P+dP.  They all heat up
    different amounts dT[i], according to their thermoelastic parameters.  Then allow them
    to equilibrate to a constant temperature dT, conserving heat within the composite.
    This works out to the formula: dT/dP = T*sum(frac[i]*Cp[i]*gr[i]/K[i])/sum(frac[i]*Cp[i])

    This function is called by :func:`burnman.geotherm.adiabatic`, and in general 
    it will not be too useful in other contexts.

    Parameters
    ----------

    pressure : float
        The pressure at which to evaluate dT/dP.
   
    temperature : float
        The temperature at which to evaluate dT/dP.

    rock : :class:`burnman.composite`
        Material for which we compute dT/dP 

    Returns
    -------
        dT/dP : float
          Adiabatic temperature gradient [K/Pa] for the composite at temperature, pressure.
    """
    top = 0
    bottom = 0
    rock.set_state(pressure, temperature)
    (fractions,minerals) = rock.unroll()
    for (fr,mineral) in zip(fractions,minerals):
        gr = mineral.grueneisen_parameter()
        K_s = mineral.adiabatic_bulk_modulus()
        C_p = mineral.heat_capacity_p()

        top += fr*gr*C_p/K_s
        bottom += fr*C_p
    
    return temperature*top/bottom


table_brown = burnman.tools.read_table("input_geotherm/brown_81.txt")
table_brown_depth = np.array(table_brown)[:,0]
table_brown_temperature = np.array(table_brown)[:,1]

table_anderson = burnman.tools.read_table("input_geotherm/anderson_82.txt")
table_anderson_depth = np.array(table_anderson)[:,0]
table_anderson_temperature = np.array(table_anderson)[:,1]




# BurnMan - a lower mantle toolkit
# Copyright (C) 2012, 2013, Heister, T., Unterborn, C., Rose, I. and Cottaar, S.
# Released under GPL v2 or later.

"""
Mineral database
"""

import Murakami_etal_2012
import Matas_etal_2007
import SLB_2011
import SLB_2011_ZSB_2013
import SLB_2005
import Murakami_2013
import other
